package com.hisschemoller.model 
{
	import com.hisschemoller.utils.audio.processors.IAudioProcessor;
	import com.hisschemoller.utils.audio.vo.AudioEventVO;

	import org.puremvc.as3.interfaces.IProxy;
	import org.puremvc.as3.patterns.proxy.Proxy;

	/**
	 * @author Wouter Hisschemöller
	 * (c) Jul 17, 2009
	 */
	public class AudioEngineProxy extends Proxy implements IProxy 
	{
		public static const NAME : String = "AudioEngineProxy";
		public static const PROCESSOR_ADDED : String = "processorAdded";
		public static const PROCESSOR_REMOVED : String = "processorRemoved";
		private var _audioEvents : Vector.<AudioEventVO> = new Vector.<AudioEventVO>();

		public function AudioEngineProxy()
		{
			super(NAME);
		}

		/**
		 * @param buffer : stereo buffer of AudioDriver.BUFFER_SIZE number of samples
		 * @param bufferStartIndex : From where in the buffer to start to write.
		 */
		public function process(buffer : Vector.<Vector.<Number>>, bufferStartIndex : uint) : void
		{
			var n : int = _audioEvents.length;
			
			while(--n > -1)
			{
				/** IAudioProcessor returns true if the sound is finished. */
				if(_audioEvents[n].audioProcessor.process(buffer, bufferStartIndex))
				{
					sendNotification(PROCESSOR_REMOVED, _audioEvents[n]);
					
					/** Create a convenient shortcut. */
					var processors : Vector.<IAudioProcessor> = _audioEvents[n].instrument.audioProcessors;
					
					/** Remove the processor from the instrument. */
					processors.splice(processors.indexOf(_audioEvents[n].audioProcessor), 1);
					
					/** Remove the event from the list here in the engine. */
					_audioEvents.splice(n, 1);
				}
			}
		}

		/**
		 * Add a new  audio event to the list of events to play.
		 */
		public function addAudioEvent(audioEventVO : AudioEventVO) : void
		{
			/** Add the processor to the instrument. */
			audioEventVO.instrument.audioProcessors.splice(0, 0, audioEventVO.audioProcessor);
			
			/** Add the event to the list here in the engine. */
			_audioEvents.splice(0, 0, audioEventVO);
			
			sendNotification(PROCESSOR_ADDED, audioEventVO);
		}

		/**
		 * 
		 */
		public function removeAllAudioEvents() : void
		{
			var n : int = _audioEvents.length;
			while(--n > -1)
			{
				_audioEvents[n].audioProcessor.stop();
			}
		}
	}
}
