package com.hisschemoller.utils.audio.instruments 
{
	import com.hisschemoller.utils.audio.processors.IAudioProcessor;
	import com.hisschemoller.utils.audio.processors.PitchedSampleProcessor;
	import com.hisschemoller.utils.audio.vo.PitchedSampleProgramVO;
	import com.hisschemoller.utils.audio.vo.ProgramVO;
	import com.hisschemoller.utils.midi.MidiEventVO;
	import com.hisschemoller.utils.midi.MidiNoteOnVO;
	import com.hisschemoller.utils.midi.MidiSpecification;	
	/**
	 * @author Wouter Hisschemoller
	 * (c) Jan 29, 2010
	 */
	public class PitchedSampleInstrument implements IAudioInstrument
	{
		private var _audioProcessors : Vector.<PitchedSampleProcessor> = new Vector.<PitchedSampleProcessor>();
		private var _programs : Vector.<PitchedSampleProgramVO> = new Vector.<PitchedSampleProgramVO>(MidiSpecification.MAX_VALUE_7BIT, true);
		private var _selectedProgram : PitchedSampleProgramVO;
		
		public function PitchedSampleInstrument()
		{
			
		}
		
		/**
		 * Add a program for this instrument.
		 */
		public function addProgram(number : uint, programVO : ProgramVO) : void
		{
			if(!(programVO is PitchedSampleProgramVO)) return;
			
			_programs[number] = PitchedSampleProgramVO(programVO);
		}
		
		/**
		 * 
		 */
		public function addProcessor(noteOnVO : MidiNoteOnVO) : IAudioProcessor
		{
			var audioProcessor : PitchedSampleProcessor = new PitchedSampleProcessor(_selectedProgram, noteOnVO.deltaTime);
			_audioProcessors.push(audioProcessor);
			
			return audioProcessor;
		}
		
		/**
		 * 
		 */
		public function removeProcessor(audioProcessor : IAudioProcessor) : void
		{
			var n : int = _audioProcessors.length;
			while(--n > -1)
			{
				if(_audioProcessors[n] == audioProcessor) _audioProcessors.splice(n, 1);
			}
		}
		
		/**
		 * 
		 */
		public function dispose() : void
		{
			
		}
		
		/**
		 * 
		 */
		public function selectProgram(number : uint) : void
		{
			_selectedProgram = _programs[number];
		}
		
		/**
		 * 
		 */
		public function get selectedProgram() : ProgramVO
		{
			return _selectedProgram;
		}
	}
}
